from .adapters import from_pydantic
from .encoders import CSVEncoder, JSONEncoder, XMLEncoder
from .extraction import create_extraction_chain
from .nodes import Number, Object, Option, Selection, Text
from .type_descriptors import (
    BulletPointDescriptor,
    TypeDescriptor,
    TypeScriptDescriptor,
)
from .version import __version__

__all__ = (
    "BulletPointDescriptor",
    "create_extraction_chain",
    "CSVEncoder",
    "from_pydantic",
    "JSONEncoder",
    "Number",
    "Object",
    "Option",
    "Selection",
    "Text",
    "TypeDescriptor",
    "TypeScriptDescriptor",
    "XMLEncoder",
    "__version__",
)



"""Type-definitions for encoders.

This file only contains the interface for encoders.

* Added a pre-built format instruction segment.
* May remove it at some point later or modify it if we discover that
  there are many ways of phrasing the format instructions.
"""
import abc
from typing import Any

from kor.nodes import AbstractSchemaNode


class Encoder(abc.ABC):
    """Abstract interface for an encoder.

    The encoder is responsible for encoding and decoding the Output
    portion of examples provided to the LLM.

    It must implement a method called get_instruction_segment that
    contains instructions for the LLM on how to format its output.
    """

    @abc.abstractmethod
    def encode(self, data: Any) -> str:
        """Encode the data."""

    @abc.abstractmethod
    def decode(self, text: str) -> Any:
        """Decode the text."""

    @abc.abstractmethod
    def get_instruction_segment(self) -> str:
        """Get the format instructions for the given decoder.

        Used to guide the LLM on how to format its output.
        """
        raise NotImplementedError()


class SchemaBasedEncoder(Encoder, abc.ABC):
    """Abstract interface for an encoder that has the data schema.

    Inherit from this encoder if the encoder needs to know the schema
    of the data that's being encoded.
    """

    def __init__(self, node: AbstractSchemaNode, **kwargs: Any) -> None:
        """Attach node to the encoder to allow the encoder to understand schema."""
        self.node = node



"""Module for code that generates examples for a given input.

At the moment, this code only has a simple implementation that concatenates all the
examples, but one may want to select or generate examples in a smarter way, or take
into account the finite size of the context window and limit the number of examples.

The code uses a default encoding of XML. This encoding should match the parser.
"""
from typing import Any, List, Tuple

from kor.nodes import (
    AbstractSchemaNode,
    AbstractVisitor,
    ExtractionSchemaNode,
    Object,
    Option,
    Selection,
    TypeVar,
)

T = TypeVar("T")


class SimpleExampleAggregator(AbstractVisitor[List[Tuple[str, str]]]):
    """Use to visit node and all of its descendents and aggregates all examples."""

    def visit_option(self, node: "Option", **kwargs: Any) -> List[Tuple[str, str]]:
        """Should not visit Options directly."""
        raise AssertionError("Should never visit an Option node.")

    @staticmethod
    def _assemble_output(node: AbstractSchemaNode, data: Any) -> Any:
        """Assemble the output data according to the type of the node."""
        if not data:
            return {}
        if node.many and not isinstance(data, (tuple, list)):
            data = [data]
        return {node.id: data}

    def visit_object(self, node: "Object", **kwargs: Any) -> List[Tuple[str, str]]:
        """Implementation of an object visitor."""
        examples = []
        if node.examples:
            object_examples = [
                # Looks like false positive from mypy
                # Can investigate how to simplify at a later point.
                (
                    example_input,
                    self._assemble_output(node, example_output),
                )
                for example_input, example_output in node.examples
            ]
            examples.extend(object_examples)

        # Collect examples from children
        for child in node.attributes:
            child_examples = child.accept(self)
            # Take care of namespaces
            child_examples = [
                (example_input, self._assemble_output(node, example_output))
                for example_input, example_output in child_examples
            ]
            examples.extend(child_examples)

        return examples

    def visit_selection(
        self, node: "Selection", **kwargs: Any
    ) -> List[Tuple[str, str]]:
        """Selection visitor."""
        examples = []
        for option in node.options:
            for example in option.examples:
                examples.append((example, self._assemble_output(node, option.id)))

        for example_input, example_output in node.examples:
            examples.append(
                (example_input, self._assemble_output(node, example_output))
            )

        for null_example in node.null_examples:
            examples.append((null_example, ""))
        return examples

    def visit_default(
        self, node: "AbstractSchemaNode", **kwargs: Any
    ) -> List[Tuple[str, str]]:
        """Default visitor implementation."""
        if not isinstance(node, ExtractionSchemaNode):
            raise AssertionError()
        examples = []

        for text, extraction in node.examples:
            value = self._assemble_output(node, extraction)
            examples.append((text, value))
        return examples

    def visit(self, node: "AbstractSchemaNode") -> List[Tuple[str, str]]:
        """Entry-point."""
        return node.accept(self)


# PUBLIC API


def generate_examples(node: AbstractSchemaNode) -> List[Tuple[str, str]]:
    """Generate examples for a given element.

    A rudimentary implementation that simply concatenates all available examples
    from the components across the entire element tree.

    Does not provide a way to impose constraints (e.g., select a subset of examples
    to meet a constraint on the overall number of tokens.)

    Args:
        node: AbstractInput

    Returns:
        list of 2-tuples containing input, output pairs
    """
    return SimpleExampleAggregator().visit(node)



from collections import defaultdict
from html.parser import HTMLParser
from typing import Any, DefaultDict, Dict, List, Mapping, Optional, Sequence, Union

from kor.encoders.typedefs import Encoder

LiteralType = Union[str, int, float]


def _write_literal(tag_name: str, value: LiteralType) -> str:
    """Write literal."""
    return f"<{tag_name}>{value}</{tag_name}>"


def _write_list(tag_name: str, values: Sequence[LiteralType]) -> str:
    """Write list."""
    return "".join(_write_tag(tag_name, value) for value in values)


def _write_dict(tag_name: str, data: Mapping[str, Any]) -> str:
    """Write a dict."""
    s_data = "".join(
        [
            _write_tag(key, value)
            for key, value in sorted(data.items(), key=lambda item: item[0])
        ]
    )
    return _write_tag(tag_name, s_data)


def _write_tag(
    tag_name: str, data: Union[LiteralType, Sequence[LiteralType], Mapping[str, Any]]
) -> str:
    """Write a tag."""
    # Dispatch based on type.
    if isinstance(data, (str, int, float)):
        return _write_literal(tag_name, data)
    elif isinstance(data, list):
        return _write_list(tag_name, data)
    elif isinstance(data, dict):
        return _write_dict(tag_name, data)
    else:
        raise NotImplementedError(f"No support for {tag_name}")


class TagParser(HTMLParser):
    def __init__(self) -> None:
        """A heavy-handed solution, but it's fast for prototyping.

        Might be re-implemented later to restrict scope to the limited grammar, and
        more efficiency.

        Uses an HTML parser to parse a limited grammar that allows
        for syntax of the form:

            INPUT -> JUNK? VALUE*
            JUNK -> JUNK_CHARACTER+
            JUNK_CHARACTER -> whitespace | ,
            VALUE -> <IDENTIFIER>DATA</IDENTIFIER> | OBJECT
            OBJECT -> <IDENTIFIER>VALUE+</IDENTIFIER>
            IDENTIFIER -> [a-Z][a-Z0-9_]*
            DATA -> .*

        Interprets the data to allow repetition of tags and recursion
        to support representation of complex types.

        ^ Just another approximately wrong grammar specification.
        """
        super().__init__()

        self.parse_data: DefaultDict[str, List[Any]] = defaultdict(list)
        self.stack: List[DefaultDict[str, List[str]]] = [self.parse_data]
        self.success = True
        self.depth = 0
        self.data: Optional[str] = None

    def handle_starttag(self, tag: str, attrs: Any) -> None:
        """Hook when a new tag is encountered."""
        self.depth += 1
        self.stack.append(defaultdict(list))
        self.data = None

    def handle_endtag(self, tag: str) -> None:
        """Hook when a tag is closed."""
        self.depth -= 1
        top_of_stack = dict(self.stack.pop(-1))  # Pop the dictionary we don't need it

        # If a lead node
        is_leaf = self.data is not None
        # Annoying to type here, code is tested, hopefully OK
        value = self.data if is_leaf else top_of_stack
        # Difficult to type this correctly with mypy (maybe impossible?)
        # Can be nested indefinitely, so requires self referencing type
        self.stack[-1][tag].append(value)  # type: ignore
        # Reset the data so we if we encounter a sequence of end tags, we
        # don't confuse an outer end tag for belonging to a leaf node.
        self.data = None

    def handle_data(self, data: str) -> None:
        """Hook when handling data."""
        # The only data that's allowed is whitespace or a comma surrounded by whitespace
        if self.depth == 0 and data.strip() not in (",", ""):
            # If this is triggered the parse should be considered invalid.
            self.success = False
        self.data = data


# PUBLIC API


class XMLEncoder(Encoder):
    """Experimental XML encoder to encode and decode data.

    .. warning::
        This encoder is not recommended for usage, at least not without further
        benchmarking for your use-case.

    The decoder re-interprets all data types as lists, which makes validating
    and using parser results more involved. It's unclear whether the encoder
    offers more advantages over other encoders (e.g., JSON or CSV).

    The encoder would encode the following dictionary

    .. code-block:: JSON

        {
            "color": ["red", "blue"],
            "height": ["6.1"],
            "width": ["3"],
        }

    As:

    .. code-block:: XML

        <color>red</color><height>6.1</height><width>3</width><color>blue</color>

    A tag be repeated multiple times to represent multiple list elements.
    """

    def encode(self, obj: Mapping[str, Any]) -> str:
        """Encode the object as XML."""
        if not isinstance(obj, dict):
            raise TypeError(f"Expected {obj} to be of type dict, got {type(obj)}")
        return "".join(_write_tag(key, value) for key, value in obj.items())

    def decode(self, text: str) -> Dict[str, List[str]]:
        """Decode the XML as an object."""
        tag_parser = TagParser()
        tag_parser.feed(text)
        if not tag_parser.success:
            return {}
        return dict(tag_parser.parse_data)

    def get_instruction_segment(self) -> str:
        """Format the instructions segment."""
        return (
            "Please enclose the extracted information in HTML style tags with the tag"
            " name corresponding to the corresponding component ID. Use angle style"
            " brackets for the tags ('>' and '<'). Only output tags when you're"
            " confident about the information that was extracted from the user's query."
            " If you can extract several pieces of relevant information from the query,"
            " then include all of them. If the type is an array, please repeat the"
            " corresponding tag name multiple times once for each relevant extraction."
            " Do NOT output anything except for the extracted information. Only output"
            " information inside the HTML style tags. Do not include any notes or any"
            " clarifications. "
        )



"""Define validator interface and provide built-in validators for common-use cases."""
import abc
from typing import Any, List, Mapping, Optional, Tuple, Type, Union

from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError


class Validator(abc.ABC):
    @abc.abstractmethod
    def clean_data(
        self, data: Union[List[Mapping[str, Any]], Mapping[str, Any]]
    ) -> Tuple[Any, List[Exception]]:
        """Validate the data and return a cleaned version of it.

        Args:
            data: the parsed data

        Returns:
            a cleaned version of the data, the type depends on the validator
        """
        raise NotImplementedError()


class PydanticValidator(Validator):
    """Use a pydantic model for validation."""

    def __init__(self, model_class: Type[BaseModel], many: bool) -> None:
        """Create a validator for a pydantic model.

        Args:
            model_class: The pydantic model class to use for validation
            many: Whether the data is expected to correspond to a single moder
                  or a list of models
        """
        self.model_class = model_class
        self.many = many

    def clean_data(
        self, data: Any
    ) -> Tuple[Union[Optional[BaseModel], List[BaseModel]], List[Exception]]:
        """Clean the data using the pydantic model.

        Args:
            data: the parsed data

        Returns:
            cleaned data instantiated as the corresponding pydantic model
        """
        if self.many:
            exceptions: List[Exception] = []
            records: List[BaseModel] = []

            for item in data:
                try:
                    records.append(self.model_class.parse_obj(item))
                except PydanticValidationError as e:
                    exceptions.append(e)
            return records, exceptions
        else:
            try:
                return self.model_class.parse_obj(data), []
            except PydanticValidationError as e:
                return None, [e]



from typing import Any, Optional, Type, Union

from langchain.chains import LLMChain
from langchain.schema import BaseLanguageModel

from kor.encoders import Encoder, InputFormatter, initialize_encoder
from kor.nodes import Object
from kor.prompts import create_langchain_prompt
from kor.type_descriptors import TypeDescriptor, initialize_type_descriptors
from kor.validators import Validator

# PUBLIC API


def create_extraction_chain(
    llm: BaseLanguageModel,
    node: Object,
    *,
    encoder_or_encoder_class: Union[Type[Encoder], Encoder, str] = "csv",
    type_descriptor: Union[TypeDescriptor, str] = "typescript",
    validator: Optional[Validator] = None,
    input_formatter: InputFormatter = None,
    **encoder_kwargs: Any,
) -> LLMChain:
    """Create an extraction chain.
    
    Args:
        llm: the language model used for extraction
        node: the schematic description of what to extract from text
        encoder_or_encoder_class: Either an encoder instance, an encoder class
                                  or a string representing the encoder class
        type_descriptor: either a TypeDescriptor or a string representing the type \
                         descriptor name
        validator: optional validator to use for validation
        input_formatter: the formatter to use for encoding the input. Used for \
                         both input examples and the text to be analyzed.
                         
            * `None`: use for single sentences or single paragraph, no formatting
            * `triple_quotes`: for long text, surround input with \"\"\"
            * `text_prefix`: for long text, triple_quote with `TEXT: ` prefix
            * `Callable`: user provided function
        encoder_kwargs: Keyword arguments to pass to the encoder class

    Returns:
        A langchain chain
        
        
    Examples:
    
    .. code-block:: python
  
        # For CSV encoding
        chain = create_extraction_chain(llm, node, encoder_or_encoder_class="csv")
    
        # For JSON encoding
        chain = create_extraction_chain(llm, node, encoder_or_encoder_class="JSON",
                                        input_formatter="triple_quotes")
    """
    if not isinstance(node, Object):
        raise ValueError(f"node must be an Object got {type(node)}")
    encoder = initialize_encoder(encoder_or_encoder_class, node, **encoder_kwargs)
    type_descriptor_to_use = initialize_type_descriptors(type_descriptor)
    return LLMChain(
        llm=llm,
        prompt=create_langchain_prompt(
            node,
            encoder,
            type_descriptor_to_use,
            validator,
            input_formatter=input_formatter,
        ),
    )



"""JSON encoder and decoder."""
import json
from typing import Any

from kor.exceptions import ParseError

from .typedefs import Encoder
from .utils import unwrap_tag, wrap_in_tag


class JSONEncoder(Encoder):
    """JSON encoder and decoder.

    The encoder by default adds additional <json> tags around the JSON output,

    Additional tags are added to the output to help identify the JSON content
    within the LLM response and extract it.

    The usage of <json> tags is similar to the usage of ```JSON and ``` marks.
    """

    def __init__(self, use_tags: bool = True) -> None:
        """Initialize the JSON encoder.

        Args:
            use_tags: Whether to wrap the output in a special JSON tags.
                      This may help identify the JSON content in cases when
                      the model attempts to add clarifying explanations.
        """
        self.use_tags = use_tags

    def encode(self, data: Any) -> str:
        """Encode the data as JSON.

        Args:
            data: JSON serializable data.

        Returns:
            The JSON encoded data as a string optionally wrapped in <json> tags.
        """
        content = json.dumps(data)
        if self.use_tags:
            return wrap_in_tag("json", json.dumps(data))
        return content

    def decode(self, text: str) -> Any:
        """Decode the text as JSON.

        If the encoder is using tags, the <json> content is identified within the text
        and then is decoded.

        Args:
            text: the text to be decoded

        Returns:
            The decoded JSON data.
        """
        if self.use_tags:
            content = unwrap_tag("json", text)
        else:
            content = text

        if content is None:
            return {}
        try:
            return json.loads(content)
        except json.JSONDecodeError as e:
            raise ParseError(e)

    def get_instruction_segment(self) -> str:
        """Get the format instructions for the given decoder.

        This is a specification to the LLM that tells it how to shape its response
        so that the response can be structured properly using the given decoder.
        """
        format_instructions = (
            "Please output the extracted information in JSON format. Do not output"
            " anything except for the extracted information. Do not add any clarifying"
            " information. Do not add any fields that are not in the schema. If the"
            " text contains attributes that do not appear in the schema, please ignore"
            " them. All output must be in JSON format and follow the schema specified"
            " above."
        )
        if self.use_tags:
            format_instructions += " Wrap the JSON in <json> tags."
        return format_instructions



"""Definitions of input elements."""
import abc
import copy
import inspect
import operator
import re
from typing import (
    Any,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

# For now, limit what's allowed for identifiers.
# The main constraints
# 1) Relying on HTML parser to parse output
# 2) One of the type descriptors is TypeScript, so we want
#    to produce valid TypeScript identifiers.
# We can lift the constraints later if it becomes important,
# not worth the effort for a v0.
VALID_IDENTIFIER_PATTERN = re.compile(r"^[a-z_][0-9a-z_]*$")

T = TypeVar("T")


def _get_all_slots(node: "AbstractSchemaNode") -> List[str]:
    """Get a list of all slots."""
    slots: List[str] = []
    for class_ in inspect.getmro(type(node)):
        if hasattr(class_, "__slots__"):
            slots.extend(class_.__slots__)
    return sorted(slots)


# Visitor is defined here for now, to avoid circular imports.
class AbstractVisitor(Generic[T], abc.ABC):
    """An abstract visitor."""

    def visit_text(self, node: "Text", **kwargs: Any) -> T:
        """Visit text node."""
        return self.visit_default(node, **kwargs)

    def visit_number(self, node: "Number", **kwargs: Any) -> T:
        """Visit text node."""
        return self.visit_default(node, **kwargs)

    def visit_object(self, node: "Object", **kwargs: Any) -> T:
        """Visit object node."""
        return self.visit_default(node, **kwargs)

    def visit_selection(self, node: "Selection", **kwargs: Any) -> T:
        """Visit selection node."""
        return self.visit_default(node, **kwargs)

    def visit_option(self, node: "Option", **kwargs: Any) -> T:
        """Visit option node."""
        return self.visit_default(node, **kwargs)

    def visit_default(self, node: "AbstractSchemaNode", **kwargs: Any) -> T:
        """Default node implementation."""
        raise NotImplementedError()


class AbstractSchemaNode(abc.ABC):
    """Abstract schema node.

    Each node is expected to have a unique ID, and should
    only use alphanumeric characters.

    The ID should be unique across all inputs that belong
    to a given form.

    The description should describe what the node represents.
    It is used during prompt generation.
    """

    __slots__ = "id", "description", "many"

    def __init__(self, *, id: str, description: str = "", many: bool = False) -> None:
        self.id = id
        self.description = description
        self.many = many

        if not VALID_IDENTIFIER_PATTERN.match(self.id):
            raise ValueError(
                f"`{self.id}` is not a valid identifier. "
                "Please only use lower cased a-z, _ or the digits 0-9"
            )

    @abc.abstractmethod
    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        raise NotImplementedError()

    # Update return type to `Self` when bumping python version.
    def replace(
        self,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        description: Optional[str] = None,
    ) -> "AbstractSchemaNode":
        """Wrapper around data-classes replace."""
        new_object = copy.copy(self)
        if id:
            new_object.id = id
        if description:
            new_object.description = description
        return new_object

    def __repr__(self) -> str:
        """Get representation of the node."""
        return f"{type(self).__name__}({self.id})"

    def __eq__(self: Any, other: Any) -> bool:
        """Equality check"""
        if not isinstance(self, AbstractSchemaNode):
            raise AssertionError(f"Cannot compare {type(self)} with {type(other)}")
        if type(self) != type(other):
            return False

        if _get_all_slots(self) == _get_all_slots(other):
            attr_getters = [operator.attrgetter(attr) for attr in self.__slots__]
            return all(getter(self) == getter(other) for getter in attr_getters)

        return False


class ExtractionSchemaNode(AbstractSchemaNode, abc.ABC):
    """An abstract definition for inputs that involve extraction.

    An extraction input can be associated with extraction examples.

    An extraction example is a 2-tuple composed of a text segment and the expected
    extraction.

    For example:

    .. code-block:: python
        [
            ("I bought this cookie for $10", "$10"),
            ("Eggs cost twelve dollars", "twelve dollars"),
        ]
    """

    __slots__ = ("examples",)

    def __init__(
        self,
        *,
        id: str,
        description: str = "",
        many: bool = False,
        examples: Sequence[Tuple[str, Union[str, Sequence[str]]]] = tuple(),
    ) -> None:
        """Initialize for extraction input."""
        super().__init__(id=id, description=description, many=many)
        self.examples = examples


class Number(ExtractionSchemaNode):
    """Built-in number input."""

    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        return visitor.visit_number(self, **kwargs)


class Text(ExtractionSchemaNode):
    """Built-in text input."""

    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        return visitor.visit_text(self, **kwargs)


class Option(AbstractSchemaNode):
    """Built-in option input must be part of a selection input."""

    __slots__ = ("examples",)

    def __init__(
        self,
        *,
        id: str,
        description: str = "",
        many: bool = False,
        examples: Sequence[str] = tuple(),
    ) -> None:
        """Initialize for extraction input."""
        if many:
            # TODO: Fix the type hierarchy so that `many` isn't provided to option.
            raise ValueError("Option inputs cannot be many.")

        super().__init__(id=id, description=description, many=many)
        self.examples = examples

    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        return visitor.visit_option(self, **kwargs)


class Selection(AbstractSchemaNode):
    """Built-in selection node (aka Enum).

    A selection input is composed of one or more options.

    A selectio node supports both examples and null_examples.

    Null examples are segments of text for which nothing should be extracted.

    Examples:

    .. code-block:: python

        selection = Selection(
            id="species",
            description="What is your favorite animal species?",
            options=[
                Option(id="dog", description="Dog"),
                Option(id="cat", description="Cat"),
                Option(id="bird", description="Bird"),
            ],
            examples=[
                ("I like dogs", "dog"),
                ("I like cats", "cat"),
                ("I like birds", "bird"),
            ],
            null_examples=[
                "I like flowers",
            ],
            many=False
        )
    """

    __slots__ = "options", "examples", "null_examples"

    def __init__(
        self,
        *,
        id: str,
        description: str = "",
        many: bool = False,
        options: Sequence[Option],
        examples: Sequence[Tuple[str, Union[str, Sequence[str]]]] = tuple(),
        null_examples: Sequence[str] = tuple(),
    ) -> None:
        """Initialize for extraction input."""
        super().__init__(id=id, description=description, many=many)

        if not options:
            raise ValueError("Selection inputs must have at least one option.")
        self.options = options
        self.examples = examples
        self.null_examples = null_examples

    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        return visitor.visit_selection(self, **kwargs)


class Object(AbstractSchemaNode):
    """Built-in representation for an object.

    Use an object node to represent an entire object that should be extracted.

    An extraction input can be associated with 2 different types of examples:

    Example:

    .. code-block:: python

        object = Object(
            id="cookie",
            description="Information about a cookie including price and name.",
            attributes=[
                Text(id="name", description="The name of the cookie"),
                Number(id="price", description="The price of the cookie"),
            ],
            examples=[
                ("I bought this Big Cookie for $10",
                    {"name": "Big Cookie", "price": "$10"}),
                ("Eggs cost twelve dollars", {}), # Not a cookie
            ],
        )

    """

    __slots__ = ("attributes", "examples")

    def __init__(
        self,
        *,
        id: str,
        description: str = "",
        many: bool = False,
        # All attributes but Option are OK.
        # May could clean up the type system to simplify this.
        attributes: Sequence[Union[ExtractionSchemaNode, Selection, "Object"]],
        examples: Sequence[
            Tuple[str, Mapping[str, Union[str, Sequence[str]]]]
        ] = tuple(),
    ) -> None:
        """Initialize for extraction input."""
        super().__init__(id=id, description=description, many=many)
        self.attributes = attributes
        self.examples = examples

    def accept(self, visitor: AbstractVisitor[T], **kwargs: Any) -> T:
        """Accept a visitor."""
        return visitor.visit_object(self, **kwargs)



from typing import Any, Callable, List, Literal, Mapping, Sequence, Tuple, Type, Union

from kor.nodes import AbstractSchemaNode

from .csv_data import CSVEncoder
from .json_data import JSONEncoder
from .typedefs import Encoder, SchemaBasedEncoder
from .xml import XMLEncoder

_ENCODER_REGISTRY: Mapping[str, Type[Encoder]] = {
    "csv": CSVEncoder,
    "xml": XMLEncoder,
    "json": JSONEncoder,
}

# Use to denote different types of formatters for the input.
InputFormatter = Union[
    Literal["text_prefix"], Literal["triple_quotes"], None, Callable[[str], str]
]


# PUBLIC API


def format_text(text: str, input_formatter: InputFormatter = None) -> str:
    """An encoder for the input text.

    Args:
        text: the text to encode
        input_formatter: the formatter to use for the input
            * None: use for single sentences or single paragraphs, no formatting
            * triple_quotes: surround input with \"\"\", use for long text
            * text_prefix: same as triple_quote but with `TEXT: ` prefix
            * Callable: user provided function

    Returns:
        The encoded text if it was encoded
    """
    if input_formatter == "text_prefix":
        return 'Text: """\n' + text + '\n"""'
    elif input_formatter == "triple_quotes":
        return '"""\n' + text + '\n"""'
    elif input_formatter is None:
        return text
    else:
        raise NotImplementedError(
            f'No support for input encoding "{input_formatter}". '
            ' Use one of "long_text" or None.'
        )


def encode_examples(
    examples: Sequence[Tuple[str, str]],
    encoder: Encoder,
    input_formatter: InputFormatter = None,
) -> List[Tuple[str, str]]:
    """Encode the output using the given encoder."""

    return [
        (
            format_text(input_example, input_formatter=input_formatter),
            encoder.encode(output_example),
        )
        for input_example, output_example in examples
    ]


def initialize_encoder(
    encoder_or_encoder_class: Union[Type[Encoder], Encoder, str],
    schema: AbstractSchemaNode,
    **kwargs: Any,
) -> Encoder:
    """Flexible way to initialize an encoder, used only for top level API.

    Args:
        encoder_or_encoder_class: Either an encoder instance, an encoder class
                                  or a string representing the encoder class.
        schema: The schema to use for the encoder.
        **kwargs: Keyword arguments to pass to the encoder class.

    Returns:
        An encoder instance
    """
    if isinstance(encoder_or_encoder_class, str):
        encoder_name = encoder_or_encoder_class.lower()
        if encoder_name not in _ENCODER_REGISTRY:
            raise ValueError(
                f"Unknown encoder {encoder_name}. "
                f"Use one of {sorted(_ENCODER_REGISTRY)}"
            )
        encoder_or_encoder_class = _ENCODER_REGISTRY[encoder_name]
    if isinstance(encoder_or_encoder_class, type(Encoder)):
        if issubclass(encoder_or_encoder_class, SchemaBasedEncoder):
            return encoder_or_encoder_class(schema, **kwargs)
        else:
            return encoder_or_encoder_class(**kwargs)
    elif isinstance(encoder_or_encoder_class, Encoder):
        if kwargs:
            raise ValueError("Unable to use kwargs with an encoder instance")
        return encoder_or_encoder_class
    else:
        raise TypeError(
            "Expected str, an encoder or encoder class, got"
            f" {type(encoder_or_encoder_class)}"
        )



"""Declare public interface for encoders.

An encoder follows the `Encoder` interface.

It can encode, decode and contains instructions about the encoding format for an LLM.
"""
from .csv_data import CSVEncoder
from .encode import InputFormatter, encode_examples, initialize_encoder
from .json_data import JSONEncoder
from .typedefs import Encoder, SchemaBasedEncoder
from .xml import XMLEncoder

__all__ = [
    "CSVEncoder",
    "encode_examples",
    "Encoder",
    "initialize_encoder",
    "InputFormatter",
    "JSONEncoder",
    "SchemaBasedEncoder",
    "XMLEncoder",
]



"""Adapters to convert from validation frameworks to Kor internal representation."""
import enum
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel

from .nodes import ExtractionSchemaNode, Number, Object, Option, Selection, Text
from .validators import PydanticValidator, Validator

# Not going to support dicts or lists since that requires recursive checks.
# May make sense to either drop the internal representation, or properly extend it
# to handle Lists, Unions etc.
# Not worth the effort, until it's clear that folks are using this functionality.
PRIMITIVE_TYPES = {str, float, int, type(None)}


def _translate_pydantic_to_kor(
    model_class: Type[BaseModel],
    *,
    name: Optional[str] = None,
    description: str = "",
    examples: Sequence[Tuple[str, Dict[str, Any]]] = tuple(),
    many: bool = False,
) -> Object:
    """Convert a pydantic model to Kor internal representation.

    Args:
        model_class: The pydantic model class to convert.
        name: The name of the model.
        description: The description of the model.
        examples: A sequence of examples of the model.
        many: Whether the model is a list of models.

    Returns:
        The Kor internal representation of the model.
    """

    attributes: List[Union[ExtractionSchemaNode, Selection, "Object"]] = []
    for field_name, field in model_class.__fields__.items():
        field_info = field.field_info
        extra = field_info.extra
        if "examples" in extra:
            field_examples = extra["examples"]
        else:
            field_examples = tuple()

        field_description = field_info.description or ""

        type_ = field.type_
        field_many = get_origin(field.outer_type_) is list
        attribute: Union[ExtractionSchemaNode, Selection, "Object"]
        # Precedence matters here since bool is a subclass of int
        if get_origin(type_) is Union:
            args = get_args(type_)

            if not all(arg in PRIMITIVE_TYPES for arg in args):
                raise NotImplementedError(
                    "Union of non-primitive types not supported. Issue with"
                    f"field: `{field_name}`. Has type: `{type_}`"
                )

            attribute = Text(
                id=field_name,
                examples=field_examples,
                description=field_description,
                many=field_many,
            )
        else:
            if issubclass(type_, BaseModel):
                attribute = _translate_pydantic_to_kor(
                    type_,
                    description=field_description,
                    examples=field_examples,
                    many=field_many,
                    name=field_name,
                )
            elif issubclass(type_, bool):
                attribute = Text(
                    id=field_name,
                    examples=field_examples,
                    description=field_description,
                    many=field_many,
                )
            elif issubclass(type_, (int, float)):
                attribute = Number(
                    id=field_name,
                    examples=field_examples,
                    description=field_description,
                    many=field_many,
                )
            elif issubclass(type_, enum.Enum):
                enum_choices = list(type_)
                attribute = Selection(
                    id=field_name,
                    description=field_description,
                    many=field_many,
                    examples=field_examples,
                    options=[Option(id=choice.value) for choice in enum_choices],
                )
            else:
                attribute = Text(
                    id=field_name,
                    examples=field_examples,
                    description=field_description,
                    many=field_many,
                )

        attributes.append(attribute)

    return Object(
        id=name or model_class.__name__.lower(),
        attributes=attributes,
        description=description,
        examples=examples,
        many=many,
    )


# PUBLIC API


def from_pydantic(
    model_class: Type[BaseModel],
    *,
    description: str = "",
    examples: Sequence[Tuple[str, Dict[str, Any]]] = tuple(),
    many: bool = False,
) -> Tuple[Object, Validator]:
    """Convert a pydantic model to Kor internal representation.

    Args:
        model_class: The pydantic model class to convert.
        description: The description of the model.
        examples: A sequence of examples to be used for the model.
        many: Whether to expect the model to be a list of models.

    Returns:
        A tuple of the Kor internal representation of the model and a validator.
    """
    schema = _translate_pydantic_to_kor(
        model_class,
        description=description,
        examples=examples,
        many=many,
    )
    validator = PydanticValidator(model_class, schema.many)
    return schema, validator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kor', 'kor.encoders']

package_data = \
{'': ['*']}

install_requires = \
['langchain>=0.0.110', 'openai>=0.27,<0.28', 'pandas>=1.5.3,<2.0.0']

setup_kwargs = {
    'name': 'kor',
    'version': '0.6.1',
    'description': 'Extract information with LLMs from text',
    'long_description': '**⚠ WARNING: Prototype with unstable API. 🚧**  \n\n[![Unit Tests](https://github.com/eyurtsev/kor/actions/workflows/test.yml/badge.svg?branch=main&event=push)](https://github.com/eyurtsev/kor/actions/workflows/test.yml)\n[![Test Docs](https://github.com/eyurtsev/kor/actions/workflows/doc_test.yaml/badge.svg?branch=main&event=push)](https://github.com/eyurtsev/kor/actions/workflows/doc_test.yaml)\n\n# Kor\n\n\nThis is a half-baked prototype that "helps" you extract structured data from text using LLMs 🧩.\n\nSpecify the schema of what should be extracted and provide some examples.\n\nKor will generate a prompt, send it to the specified LLM and parse out the\noutput. \n\nYou might even get results back.\n\nSee [documentation](https://eyurtsev.github.io/kor/).\n\n## Version >=0.4.0\n\n* Integrated with langchain framework.\n* The code below uses Kor style schema, but you can also use [pydantic](https://eyurtsev.github.io/kor/validation.html).\n\n\n```python\n\n  from langchain.chat_models import ChatOpenAI\n  from kor import create_extraction_chain, Object, Text, Number\n\n  llm = ChatOpenAI(\n      model_name="gpt-3.5-turbo",\n      temperature=0,\n      max_tokens=2000,\n      frequency_penalty=0,\n      presence_penalty=0,\n      top_p=1.0,\n  )\n\n  schema = Object(\n    id="player",\n    description=(\n        "User is controling a music player to select songs, pause or start them or play"\n        " music by a particular artist."\n    ),\n    attributes=[\n        Text(\n            id="song",\n            description="User wants to play this song",\n            examples=[],\n            many=True,\n        ),\n        Text(\n            id="album",\n            description="User wants to play this album",\n            examples=[],\n            many=True,\n        ),\n        Text(\n            id="artist",\n            description="Music by the given artist",\n            examples=[("Songs by paul simon", "paul simon")],\n            many=True,\n        ),\n        Text(\n            id="action",\n            description="Action to take one of: `play`, `stop`, `next`, `previous`.",\n            examples=[\n                ("Please stop the music", "stop"),\n                ("play something", "play"),\n                ("play a song", "play"),\n                ("next song", "next"),\n            ],\n        ),\n    ],\n    many=False,\n)\n\n  chain = create_extraction_chain(llm, schema, encoder_or_encoder_class=\'json\')\n  chain.predict_and_parse(text="play songs by paul simon and led zeppelin and the doors")[\'data\']\n```\n\n```python\n  {\'player\': {\'artist\': [\'paul simon\', \'led zeppelin\', \'the doors\']}}\n```\n\n## Compatibility\n\n`Kor` is tested against python 3.8, 3.9, 3.10, 3.11.\n\n## Installaton \n\n```sh\npip install kor\n```\n\n## 💡 Ideas\n\nIdeas of some things that could be done with Kor.\n\n* Extract data from text that matches an extraction schema.\n* Power an AI assistant with skills by precisely understanding a user request.\n* Provide natural language access to an existing API.\n\n## 🚧 Prototype\n\nPrototype! So the API is not expected to be stable!\n\n##  ✨ What does Kor excel at?  🌟 \n\n* Making mistakes! Plenty of them!\n* Slow! It uses large prompts with examples, and works best with the larger slower LLMs.\n* Crashing for long enough pieces of text! Context length window could become\n  limiting when working with large forms or long text inputs.\n\nThe expectation is that as LLMs improve some of these issues will be mitigated.\n\n## Limtations\n\nNo limitations whatsoever. Do take a look at the section directly above as well\nas at the section about compatibility.\n\n## Potential Changes\n\n* Adding validators\n* Built-in components to quickly assemble schema with examples\n* Add routing layer to select appropriate extraction schema for a use case when\n  many schema exist\n\n## 🎶 Why the name?\n\nFast to type and sufficiently unique.\n\n## Contributing\n\nIf you have any ideas or feature requests, please open an issue and share!\n\nSee [CONTRIBUTING.md](https://github.com/eyurtsev/kor/blob/main/CONTRIBUTING.md) for more information.\n\n## Other packages\n\nProbabilistically speaking this package is unlikely to work for your use case.\n\nSo here are some great alternatives:\n\n* [Promptify](https://github.com/promptslab/Promptify)\n* [MiniChain](https://srush.github.io/MiniChain/examples/stats/)\n',
    'author': 'Eugene Yurtsev',
    'author_email': 'eyurtsev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.github.com/eyurtsev/kor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



"""Code to dynamically generate appropriate LLM prompts."""
from __future__ import annotations

from typing import Any, List, Optional, Tuple

from langchain import BasePromptTemplate
from langchain.schema import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    PromptValue,
    SystemMessage,
)
from pydantic import Extra

from kor.encoders import Encoder
from kor.encoders.encode import InputFormatter, encode_examples, format_text
from kor.encoders.parser import KorParser
from kor.examples import generate_examples
from kor.nodes import Object
from kor.type_descriptors import TypeDescriptor

from .validators import Validator


class ExtractionPromptValue(PromptValue):
    """Integration with langchain prompt format."""

    text: str
    encoder: Encoder
    node: Object
    type_descriptor: TypeDescriptor
    input_formatter: InputFormatter
    prefix: str = (
        "Your goal is to extract structured information from the user's input that"
        " matches the form described below. When extracting information please make"
        " sure it matches the type information exactly. Do not add any attributes that"
        " do not appear in the schema shown below."
    )

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def get_formatted_text(self) -> str:
        """Get the text encoded if needed."""
        return format_text(self.text, input_formatter=self.input_formatter)

    def to_string(self) -> str:
        """Format the template to a string."""
        instruction_segment = self.generate_instruction_segment(self.node)
        encoded_examples = self.generate_encoded_examples(self.node)
        formatted_examples: List[str] = []

        for in_example, output in encoded_examples:
            formatted_examples.extend(
                [
                    f"Input: {in_example}",
                    f"Output: {output}",
                ]
            )

        text = self.get_formatted_text()
        formatted_examples.append(f"Input: {text}\nOutput:")
        input_output_block = "\n".join(formatted_examples)
        return f"{instruction_segment}\n\n{input_output_block}"

    def to_messages(self) -> List[BaseMessage]:
        """Format the template to chat messages."""
        instruction_segment = self.generate_instruction_segment(self.node)

        messages: List[BaseMessage] = [SystemMessage(content=instruction_segment)]
        encoded_examples = self.generate_encoded_examples(self.node)

        for example_input, example_output in encoded_examples:
            messages.extend(
                [
                    HumanMessage(content=example_input),
                    AIMessage(content=example_output),
                ]
            )

        content = self.get_formatted_text()
        messages.append(HumanMessage(content=content))
        return messages

    def generate_encoded_examples(self, node: Object) -> List[Tuple[str, str]]:
        """Generate encoded examples."""
        examples = generate_examples(node)
        return encode_examples(
            examples, self.encoder, input_formatter=self.input_formatter
        )

    def generate_instruction_segment(self, node: Object) -> str:
        """Generate the instruction segment of the extraction."""
        type_description = self.type_descriptor.describe(node)
        instruction_segment = self.encoder.get_instruction_segment()
        return f"{self.prefix}\n\n{type_description}\n\n{instruction_segment}"


class ExtractionPromptTemplate(BasePromptTemplate):
    """Extraction prompt template."""

    encoder: Encoder
    node: Object
    type_descriptor: TypeDescriptor
    input_formatter: InputFormatter

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def format_prompt(  # type: ignore[override]
        self, text: str, **kwargs: Any
    ) -> PromptValue:
        """Format the prompt."""
        return ExtractionPromptValue(
            text=text,
            encoder=self.encoder,
            node=self.node,
            type_descriptor=self.type_descriptor,
            input_formatter=self.input_formatter,
        )

    def format(self, **kwargs: Any) -> str:
        """Implementation of deprecated format method."""
        raise NotImplementedError()

    @property
    def _prompt_type(self) -> str:
        """Prompt type."""
        return "ExtractionPromptTemplate"


# PUBLIC API


def create_langchain_prompt(
    schema: Object,
    encoder: Encoder,
    type_descriptor: TypeDescriptor,
    validator: Optional[Validator] = None,
    input_formatter: InputFormatter = None,
) -> ExtractionPromptTemplate:
    """Create a langchain style prompt with specified encoder."""
    return ExtractionPromptTemplate(
        input_variables=["text"],
        output_parser=KorParser(encoder=encoder, validator=validator, schema_=schema),
        encoder=encoder,
        node=schema,
        input_formatter=input_formatter,
        type_descriptor=type_descriptor,
    )



from __future__ import annotations

from typing import Any, Dict, Optional

from pydantic import Extra

from kor.encoders import Encoder
from kor.exceptions import ParseError
from kor.nodes import Object
from kor.validators import Validator

try:
    from langchain.output_parsers.base import BaseOutputParser
except ImportError:
    from langchain.schema import BaseOutputParser  # type: ignore


class KorParser(BaseOutputParser):
    """A Kor langchain parser integration.

    This parser can use any of Kor's encoders to support encoding/decoding
    different data formats.
    """

    encoder: Encoder
    schema_: Object
    validator: Optional[Validator] = None

    @property
    def _type(self) -> str:
        """Declare the type property."""
        return "KorEncoder"

    def parse(self, text: str) -> Dict[str, Any]:
        """Parse the text."""
        try:
            data = self.encoder.decode(text)
        except ParseError as e:
            return {"data": {}, "raw": text, "errors": [e], "validated_data": {}}

        key_id = self.schema_.id

        if key_id not in data:
            if data:  # We got something parsed, but it doesn't match the schema.
                errors = [
                    ParseError(
                        "The LLM has returned structured data which does not match the"
                        " expected schema. Providing additional examples may help"
                        " improve the parse."
                    )
                ]
            else:
                errors = []
            return {"data": {}, "raw": text, "errors": errors, "validated_data": {}}

        obj_data = data[key_id]

        if self.validator:
            validated_data, exceptions = self.validator.clean_data(obj_data)
        else:
            validated_data, exceptions = {}, []

        return {
            "data": data,
            "raw": text,
            "errors": exceptions,
            "validated_data": validated_data,
        }

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True



class KorException(Exception):
    """Base class for all Kor exceptions."""


class ParseError(KorException):
    """Exception for parsing errors."""


class ValidationError(KorException):
    """Exception for validators."""



"""Code that takes an Object schema and outputs a string that describes its schema.

Without fine-tuning the LLM, the quality of the response may end up depending
on details such as the schema description in the prompt.

Users can implement their own type descriptors or customize an existing one
using inheritance and over-loading and provide the type-descriptors to
the create_extraction_chain function.
"""
import abc
from typing import Any, Iterable, List, TypeVar, Union

from kor.nodes import (
    AbstractSchemaNode,
    AbstractVisitor,
    Number,
    Object,
    Selection,
    Text,
)

T = TypeVar("T")

# PUBLIC API


class TypeDescriptor(AbstractVisitor[T], abc.ABC):
    """Abstract interface for a type-descriptor.

    A type-descriptor is responsible for taking in a schema and outputting its type
    as a string. The description is used to help the LLM generate structured output.

    A type-descriptor is a visitor that can be used to traverse the schema recursively.
    """

    @abc.abstractmethod
    def describe(self, node: Object) -> str:
        """Take in node and describe its type as a string."""
        raise NotImplementedError()


class BulletPointDescriptor(TypeDescriptor[Iterable[str]]):
    """Generate a bullet point style schema description."""

    def visit_default(self, node: "AbstractSchemaNode", **kwargs: Any) -> List[str]:
        """Default action for a node."""
        depth = kwargs["depth"]
        space = "* " + depth * " "
        return [f"{space}{node.id}: {node.__class__.__name__} # {node.description}"]

    def visit_object(self, node: Object, **kwargs: Any) -> List[str]:
        """Visit an object node."""
        depth = kwargs["depth"]
        code_lines = self.visit_default(node, depth=depth)
        for child in node.attributes:
            code_lines.extend(child.accept(self, depth=depth + 1))
        return code_lines

    def describe(self, node: Object) -> str:
        """Describe the type of the given node."""
        code_lines = node.accept(self, depth=0)
        return "\n".join(code_lines)


class TypeScriptDescriptor(TypeDescriptor[Iterable[str]]):
    """Generate a typescript style schema description."""

    def visit_default(self, node: "AbstractSchemaNode", **kwargs: Any) -> List[str]:
        """Default action for a node."""
        depth = kwargs["depth"]
        space = depth * " "

        if isinstance(node, Selection):
            finalized_type = " | ".join('"' + s.id + '"' for s in node.options)
        elif isinstance(node, Text):
            finalized_type = "string"
        elif isinstance(node, Number):
            finalized_type = "number"
        else:
            raise NotImplementedError()

        if node.many:
            finalized_type = "Array<" + finalized_type + ">"

        return [f"{space}{node.id}: {finalized_type} // {node.description}"]

    def visit_object(self, node: Object, **kwargs: Any) -> List[str]:
        """Visit an object node."""
        depth = kwargs["depth"]
        space = depth * " "

        if node.many:
            many_formatter = "Array<"
        else:
            many_formatter = ""

        code_lines = [f"{space}{node.id}: {many_formatter}{{ // {node.description}"]

        for child in node.attributes:
            code_lines.extend(child.accept(self, depth=depth + 1))

        if node.many:
            many_formatter = ">"
        else:
            many_formatter = ""

        code_lines.append(f"{space}}}{many_formatter}")
        return code_lines

    def describe(self, node: "Object") -> str:
        """Describe the node type in TypeScript notation."""
        if not isinstance(node, Object):
            raise TypeError(f"Expecting an Object node got {node}")

        code_lines = node.accept(self, depth=0)
        code = "\n".join(code_lines)
        return f"```TypeScript\n\n{code}\n```\n"


def initialize_type_descriptors(
    type_descriptor: Union[TypeDescriptor, str]
) -> TypeDescriptor:
    """Initialize the type descriptors."""
    if isinstance(type_descriptor, str):
        if type_descriptor == "bullet_point":
            return BulletPointDescriptor()
        elif type_descriptor == "typescript":
            return TypeScriptDescriptor()
        else:
            raise ValueError(
                f"Unknown type descriptor: {type_descriptor}. Use one of: bullet_point,"
                " typescript or else provide an instance of TypeDescriptor."
            )
    return type_descriptor
